# Copyright 2022, Gaurav Juvekar
# SPDX-License-Identifier: MIT
import base64


class XOauth2(object):
    def __init__(self, user, token):
        self.xoauth2_string = f'user={user}\x01auth=Bearer {token}\x01\x01'

    def authenticate_imap(self):
        return lambda _: self.xoauth2_string



# Copyright 2022, Gaurav Juvekar
# SPDX-License-Identifier: MIT
import datetime
import itertools
import time
import typing

from . import types


def pipeline(message, actions):
    actions = iter(actions)
    while True:
        try:
            action = next(actions)
        except StopIteration:
            break

        action.remote = message.remote
        try:
            further = action(message)
            if further is None:
                raise Exception(msg=f'Action: {action} returned None')
        except StopIteration:
            return
        actions = itertools.chain(further, actions)


def start(remote,
          dir_actions: typing.Dict[types.Directory, typing.List['Action']] = dict(),
          interval=datetime.timedelta(seconds=5),
          count=float('inf')):
    """
    Start applying :class:`~.action.Action` s to all messages in specified
    directories.

    :param Dict[Directory, List[Action]] dir_actions:
       :class:`~.action.Action` to apply to all messages in the directory
    :param interval: duration to wait after each pass
    :param count: number of times to loop through all directories
    """
    while count > 0:
        for dir_ in remote.list_dirs():
            if dir_ in dir_actions:
                for message in remote.get_messages(dir_):
                    pipeline(message, dir_actions[dir_])
        count -= 1

        time.sleep(interval.seconds)



# Copyright 2022, Gaurav Juvekar
# SPDX-License-Identifier: MIT
import abc
import logging
import typing

import imapclient

from . import message, types

log = logging.getLogger(__name__)

imapclient.imaplib.Debug = 0


class Remote(abc.ABC):
    @abc.abstractmethod
    def list_dirs(self) -> typing.Iterable[types.Directory]:
        """
        List all ``Directory`` in the mailbox.
        """
        pass

    @abc.abstractmethod
    def list_messages(self, dir_: types.Directory) -> types.Uid:
        """
        List unique identifiers for all messages in ``dir_``. These identifiers
        must be unique for the entire mailbox.
        """
        pass

    @abc.abstractmethod
    def fetch_envelope(self, msg_id: types.Uid):
        """
        Fetch the envelope parsed from the email headers.
        """
        pass

    @abc.abstractmethod
    def fetch_body(self, msg_id: types.Uid):
        """
        Fetch the full email body.
        """
        pass

    def get_messages(self, dir_: types.Directory) -> typing.Iterable[message.Message]:
        """
        Get all messages in ``dir_``
        """
        for msg_id in self.list_messages(dir_):
            envelope = self.fetch_envelope(msg_id)
            yield message.Message(uid=msg_id, envelope=envelope,
                                  dir_=dir_, remote=self)

    @abc.abstractmethod
    def move_message_id(self, msg_id: types.Uid, target_dir: types.Directory) -> types.Uid:
        """
        Move ``msg_id`` to ``target_dir``.
        """
        pass

    def move_message(self, msg: message.Message, target_dir: types.Directory):
        """
        Move ``msg`` to ``taget_dir``.
        """
        new_uid = self.move_message_id(msg.uid, target_dir)
        msg.dir_ = target_dir
        msg.uid = new_uid


class Imap(Remote):
    def __init__(self, host, user, token, **kwargs):
        super().__init__(**kwargs)
        self.connection = imapclient.IMAPClient(host)
        self.connection.oauth2_login(user, access_token=token)

    def list_dirs(self):
        for flags, delim, name in self.connection.list_folders():
            name_components = tuple(name.split(delim.decode()))
            yield name_components

    def list_messages(self, dir_):
        self.connection.select_folder('/'.join(dir_))
        for uid in self.connection.search():
            # IMAP message uid are unique only within the directory. Create a
            # composite uid that contains the directory.
            yield (dir_, uid)

    def fetch_envelope(self, msg_id):
        dir_, uid = msg_id
        self.connection.select_folder('/'.join(dir_))
        ret = self.connection.fetch(uid, ['UID', 'ENVELOPE'])
        msg = ret[uid]
        return msg[b'ENVELOPE']

    def fetch_body(self, msg_id):
        dir_, uid = msg_id
        self.connection.select_folder('/'.join(dir_))
        ret = self.connection.fetch(uid, ['UID', 'RFC822'])
        msg = ret[uid]
        return msg[b'RFC822']

    def move_message_id(self, msg_id, target_dir):
        dir_, uid = msg_id
        self.connection.select_folder('/'.join(dir_))
        self.connection.move([uid], '/'.join(target_dir))
        return (target_dir, uid)



# Copyright 2022, Gaurav Juvekar
# SPDX-License-Identifier: MIT
import collections
import re
import typing


class Address(collections.namedtuple('Address', 'name mailbox host',
                                     defaults=(None, None, None))):
    @classmethod
    def from_imapclient(cls, addr):
        return cls(name=addr.name, mailbox=addr.mailbox, host=addr.host)

    def re_match(self, addr):
        sname, smbox, shost = self
        if sname is None:
            sname = b''

        if any(x is None for x in (addr.name, addr.mailbox, addr.host)):
            return False
        return all((re.fullmatch(addr.name, sname),
                    re.fullmatch(addr.mailbox, smbox),
                    re.fullmatch(addr.host, shost)))


"""
A directory on the mail server made up of the path components of the directory
"""
Directory = typing.Tuple[str]

Uid = typing.Hashable



# Copyright 2022, Gaurav Juvekar
# SPDX-License-Identifier: MIT
"""
Basic building blocks for doing stuff with a
:class:`~remote_email_filtering.message.Message`
"""
import abc
import logging

log = logging.getLogger(__name__)


class Action(abc.ABC):
    """
    A callable that does something with a :class:`~.message.Message`

    Instances will be called with :class:`~.message.Message` instances.
    The :attr:`remote` attribute will be set to a :class:`Remote` before
    calling.

    """

    def __init__(self):
        self.remote = None

    @abc.abstractmethod
    def __call__(self, msg) -> 'Iterable[Action]':
        """
        When called, an :class:`Action` should return an iterable of other
        :class:`Action` s that will be applied to the ``msg`` being processed.
        """
        pass


class Stop(Action):
    """
    Stop processing any futher :class:`Action` for the current
    :class:`~.message.Message`
    """

    def __call__(self, msg):
        raise StopIteration()


class Move(Action):
    def __init__(self, destination: tuple[str]):
        """
        Move the :class:`~.message.Message` to ``destination`` directory.

        :param tuple[str] destination: the destination directory on the server
        """
        super().__init__()
        self.destination = destination

    def __call__(self, msg):
        log.info(f'Moving {msg.dir_}/{msg.Subject} to {target_dir}')
        self.remote.move_message(msg, self.destination)
        return []



# Copyright 2022, Gaurav Juvekar
# SPDX-License-Identifier: MIT
import email
import email.header
import email.policy
import typing

from . import types


class Message(object):
    """
    An email message with convenient properties
    """
    def __init__(self, uid: types.Uid,
                 envelope, remote, dir_=None, rfc822_bytes=None):
        """
        :param uid: A unique identifier for a message within ``dir_``
        :param envelope: The envelope structure parsed from headers
        :param remote: A :class:`~.remote.Remote` used to lazy-load the body
        :param tuple[str] dir_: the mailbox directory that this email is in
        """
        self.uid = uid
        self.envelope = envelope._asdict()
        for field in ('cc', 'bcc', 'from_', 'reply_to', 'sender', 'to'):
            if not self.envelope[field]:
                self.envelope[field] = []
            self.envelope[field] = tuple((types.Address.from_imapclient(x)
                                          for x in self.envelope[field]))

        self.remote = remote
        self.dir_ = dir_
        self.raw = rfc822_bytes
        self._body = None
        if rfc822_bytes is not None:
            self._body = email.message_from_bytes(self.raw,
                                                  policy=email.policy.default)

    @property
    def body(self):
        if self._body is None:
            self.raw = self.remote.fetch_body(self.uid)
            self._body = email.message_from_bytes(self.raw,
                                                  policy=email.policy.default)
        return self._body

    @property
    def To(self):
        return self.envelope['to']

    @property
    def Cc(self):
        return self.envelope['cc']

    @property
    def From(self):
        return self.envelope['from_']

    @property
    def Recipients(self):
        return self.To + self.Cc

    @property
    def Subject(self):
        return self.envelope['subject']

    @property
    def SaneSubject(self):
        ascii_header = self.Subject.decode('ascii')
        encoded, charset = email.header.decode_header(ascii_header)[0]
        if charset is not None:
            ret = encoded.decode(charset, errors='replace')
        else:
            ret = encoded
        return ret
